package org.ten60.netkernel.mod.pds;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer1.representation.DeterminateStringRep;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;


public class PDSH2Impl extends StandardAccessorImpl
{	
	private static final String SQL_SOURCE_SET =
		"SELECT "+
		"RESOURCES.ID, RESOURCES.INSTANCE, RESOURCES.ZONE, RESOURCES.IDENTIFIER "+
		"FROM RESOURCES,SETMEMBERSHIPS "+
		"WHERE "+
		"(	RESOURCES.ID=SETMEMBERSHIPS.RESOURCEID AND "+
		"SETMEMBERSHIPS.SETID='%s' "+ 
		");";
	
	private static final String SQL_SOURCE =
		"SELECT DATA FROM RESOURCES,REPRESENTATIONS WHERE "+
		"RESOURCES.ID=REPRESENTATIONS.RESOURCEID AND "+
		"RESOURCES.INSTANCE='%s' AND "+
		"RESOURCES.ZONE='%s' AND "+
		"RESOURCES.IDENTIFIER='%s' ; ";	
	
	private static final String SQL_EXISTS=
		"SELECT ID FROM resources WHERE "+
		"resources.instance='%s' AND "+
		"resources.zone='%s' AND "+
		"resources.identifier='%s'	; ";	
		
	
	public PDSH2Impl()
	{	this.declareThreadSafe();
		//this.declareUnhandledExceptionsExpired(false);
	}

	@Override
	public void onDelete(INKFRequestContext context) throws Exception
	{	Args args=new Args(context);
		Object configuration=context.source("active:pdsDynamicRDBMSConfig");
		String resourceid=getResourceID(context, args, configuration);
		
		if(resourceid!=null)
		{	String sqlref="res:/org/ten60/netkernel/mod/pds/sql/deleteSQL.xml";
			if(args.isIdentifierSetBoundary())
			{	sqlref="res:/org/ten60/netkernel/mod/pds/sql/deleteSetSQL.xml";
			}
			INKFRequest req=context.createRequest("active:freemarker");
			req.addArgument("operator", sqlref);
			req.addArgumentByValue("id", resourceid);
			req.setRepresentationClass(String.class);
			String s=(String)context.issueRequest(req);
			
			req=context.createRequest("active:sqlBatch");
			req.addArgumentByValue("operand", s);
			req.addArgumentByValue("configuration",configuration);
			context.issueRequest(req);
		}

		INKFRequest req=context.createRequest("active:cutGoldenThread");
		req.addArgument("id", PDSUtils.getGoldenThread(args.mInstance, args.mZone, args.mIdentifier));
		req.addArgument("id", PDSUtils.getGoldenThread(args.mInstance, args.mZone, args.getSetIdentifier()));
		context.issueRequest(req);
		
		INKFResponse resp=context.createResponseFrom(resourceid!=null);
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
	}

	@Override
	public void onExists(INKFRequestContext context) throws Exception
	{	Args args=new Args(context);
		Object configuration=context.source("active:pdsDynamicRDBMSConfig");
		boolean rep=getResourceID(context, args, configuration)!=null;
		INKFRequest req=context.createRequest("active:attachGoldenThread");
		req.addArgument("id", PDSUtils.getGoldenThread(args.mInstance, args.mZone, args.mIdentifier));
		req.addArgument("id", PDSUtils.getGoldenThread(args.mInstance, args.mZone, args.getSetIdentifier()));
		context.issueRequest(req);
		context.createResponseFrom(rep);
	}

	
	@Override
	public synchronized void onSink(INKFRequestContext context) throws Exception
	{	Args args=new Args(context);
		if(args.isIdentifierSetBoundary())
		{	throw new NKFException("Sets do not support SINK");			
		}
		Object configuration=context.source("active:pdsDynamicRDBMSConfig");
		String resourceID=getResourceID(context, args, configuration);
		String blobUpdateSQL=null;
		String setGoldenThread=null;
		
		if(resourceID==null)
		{	INKFRequest req=context.createRequest("active:sqlQuery");
			req.addArgument("operand", "res:/org/ten60/netkernel/mod/pds/sql/nextIDSQL.txt");
			req.addArgument("configuration","active:pdsDynamicRDBMSConfig");
			req.setRepresentationClass(IHDSNode.class);
			IHDSNode n=(IHDSNode)context.issueRequest(req);
			resourceID=n.getFirstValue("resultset/row/ID").toString();
			
			req=context.createRequest("active:freemarker");
			req.addArgument("operator", "res:/org/ten60/netkernel/mod/pds/sql/sinkInsertBlobSQL.txt");
			req.addArgumentByValue("args", new Object[]{args, resourceID});
			req.setRepresentationClass(String.class);
			blobUpdateSQL=(String)context.issueRequest(req);
			
			//DB Update Representations
			req=context.createRequest("active:sqlUpdateBlob");
			req.addArgumentByValue("operand", new DeterminateStringRep(blobUpdateSQL));
			req.addArgumentByValue("configuration",configuration);
			req.addArgumentByValue("param", context.sourcePrimary(IReadableBinaryStreamRepresentation.class));
			context.issueRequest(req);
			
			//Deal with Resources
			req=context.createRequest("active:freemarker");
			req.addArgument("operator", "res:/org/ten60/netkernel/mod/pds/sql/sinkSQL.txt");
			req.addArgumentByValue("args",  new Object[]{args, resourceID});
			req.setRepresentationClass(String.class);
			String s=(String)context.issueRequest(req);
			
			//Prepare Batch
			HDSBuilder b=new HDSBuilder();
			b.pushNode("batch");
			b.addNode("sql", s);
			
			//Deal with Set
			Args setArgs=new Args(args.mInstance, args.mZone, args.getSetIdentifier());
			String setID=getResourceID(context, setArgs, configuration);
			if(setID==null)
			{	req=context.createRequest("active:sqlQuery");
				req.addArgument("operand", "res:/org/ten60/netkernel/mod/pds/sql/nextIDSQL.txt");
				req.addArgument("configuration","active:pdsDynamicRDBMSConfig");
				req.setRepresentationClass(IHDSNode.class);
				n=(IHDSNode)context.issueRequest(req);
				setID=n.getFirstValue("resultset/row/ID").toString();
				
				req=context.createRequest("active:freemarker");
				req.addArgument("operator", "res:/org/ten60/netkernel/mod/pds/sql/sinkSQL.txt");
				req.addArgumentByValue("args",  new Object[]{setArgs, setID});
				req.setRepresentationClass(String.class);
				s=(String)context.issueRequest(req);
				b.addNode("sql", s);
			}
			req=context.createRequest("active:freemarker");
			req.addArgument("operator", "res:/org/ten60/netkernel/mod/pds/sql/sinkResourceSetsSQL.txt");
			req.addArgumentByValue("args",  new Object[]{setID, resourceID});
			req.setRepresentationClass(String.class);
			s=(String)context.issueRequest(req);
			b.addNode("sql", s);
			
			b.popNode();
			IHDSNode sqlBatch=b.getRoot();
			
			//DB Update Resources and Sets
			req=context.createRequest("active:sqlBatch");
			req.addArgumentByValue("operand", sqlBatch);
			req.addArgumentByValue("configuration",configuration);
			context.issueRequest(req);
			
			//Cut the Set Golden Thread
			setGoldenThread=PDSUtils.getGoldenThread(setArgs.mInstance, setArgs.mZone, setArgs.mIdentifier);
		}
		else
		{	INKFRequest req=context.createRequest("active:freemarker");
			req.addArgument("operator", "res:/org/ten60/netkernel/mod/pds/sql/sinkUpdateBlobSQL.txt");
			req.addArgumentByValue("resourceid", resourceID);
			req.setRepresentationClass(String.class);
			blobUpdateSQL=(String)context.issueRequest(req);
			
			//DB Update Representations
			req=context.createRequest("active:sqlUpdateBlob");
			req.addArgumentByValue("operand",  new DeterminateStringRep(blobUpdateSQL));
			req.addArgumentByValue("configuration",configuration);
			req.addArgumentByValue("param", context.sourcePrimary(IReadableBinaryStreamRepresentation.class));
			context.issueRequest(req);
		}

		INKFRequest req=context.createRequest("active:cutGoldenThread");
		req.addArgument("id", PDSUtils.getGoldenThread(args.mInstance, args.mZone, args.mIdentifier));
		if(setGoldenThread!=null)
		{	req.addArgument("id", setGoldenThread);
		}
		context.issueRequest(req);
	}

	@Override
	public void onSource(INKFRequestContext context) throws Exception
	{	Args args=new Args(context);
		if(args.isIdentifierSetBoundary())
		{	Object configuration=context.source("active:pdsDynamicRDBMSConfig");
			String setID=getResourceID(context, args, configuration);
			Object rep=null;
			HDSBuilder b = new HDSBuilder();
			b.pushNode("set");
			if(setID!=null)
			{
				String sql=String.format(SQL_SOURCE_SET, setID, setID);

				try
				{	INKFRequest req=context.createRequest("active:sqlQuery");
					req.addArgumentByValue("operand", new DeterminateStringRep(sql));
					req.addArgumentByValue("configuration",configuration);
					req.setRepresentationClass(IHDSNode.class);
					IHDSNode set=(IHDSNode)context.issueRequest(req);
					for (IHDSNode row : set.getNodes("/resultset/row"))
					{	b.addNode("identifier", row.getFirstValue("IDENTIFIER"));
					}
				}
				catch(Exception e)
				{	/*Do nothing*/
					throw e;
				}
			}
			rep=b.getRoot();
			INKFRequest req=context.createRequest("active:attachGoldenThread");
			req.addArgument("id", PDSUtils.getGoldenThread(args.mInstance, args.mZone, args.mIdentifier));
			req.addArgument("id", PDSUtils.getGoldenThread(args.mInstance, args.mZone, args.getSetIdentifier()));
			context.issueRequest(req);
			context.createResponseFrom(rep);
		}
		else
		{
			String sql=String.format(SQL_SOURCE, args.getInstance(), args.getZone(), args.getIdentifier());
			
			Object rep=null;
			try
			{	Object configuration=context.source("active:pdsDynamicRDBMSConfig");
				INKFRequest req=context.createRequest("active:sqlQueryBlob");
				req.addArgumentByValue("operand", new DeterminateStringRep(sql));
				req.addArgumentByValue("configuration",configuration);
				rep=context.issueRequest(req);
			}
			catch(Exception e)
			{	//throw context.createFormattedException("EX_PDS_NOT_FOUND", "MSG_PDS_NOT_FOUND", null, null,  args.mIdentifier);
				rep=null;
			}
			INKFRequest req=context.createRequest("active:attachGoldenThread");
			req.addArgument("id", PDSUtils.getGoldenThread(args.mInstance, args.mZone, args.mIdentifier));
			context.issueRequest(req);
			context.createResponseFrom(rep);
		}
	}
	
	private String getResourceID(INKFRequestContext context, Args args, Object configuration) throws NKFException
	{	
		String sql=String.format(SQL_EXISTS, args.getInstance(), args.getZone(), args.getIdentifier());

		INKFRequest req=context.createRequest("active:sqlQuery");
		req.addArgumentByValue("operand", new DeterminateStringRep(sql));
		req.addArgumentByValue("configuration", configuration);
		req.setRepresentationClass(IHDSNode.class);
		IHDSNode n=(IHDSNode)context.issueRequest(req);
		
		String resourceid=null;
		Object rid=n.getFirstValue("resultset/row/ID");
		if(rid!=null)
		{	resourceid=rid.toString();			
		}
		return resourceid;
	}
	
	public static class Args
	{	public String mInstance;
		public String mZone;
		public String mIdentifier;
		
		private Args(INKFRequestContext context)
		{	mInstance=context.getThisRequest().getArgumentValue("instance");
			mZone=context.getThisRequest().getArgumentValue("zone");
			mIdentifier=context.getThisRequest().getArgumentValue("pds");
		}
		private Args(String aInstance, String aZone, String aIdentifier)
		{	mInstance=aInstance;
			mZone=aZone;
			mIdentifier=aIdentifier;
		}
		
		public String getInstance()
		{	return mInstance;			
		}
		public String getZone()
		{	return mZone;			
		}
		public String getIdentifier()
		{	return mIdentifier;			
		}
		public boolean isIdentifierSetBoundary()
		{	return PDSAccessor.isIdentifierSetBoundary(mIdentifier);	
		}
		public String getSetIdentifier()
		{	return PDSAccessor.getSetIdentifier(mIdentifier);
		}
		
		@Override
		public boolean equals(Object obj)
		{	boolean result=false;
			if (obj instanceof Args)
			{	Args other=(Args)obj;
				result=mIdentifier.equals(other.mIdentifier) && mZone.equals(other.mZone) && mInstance.equals(other.mInstance);
			}
			return result;			
		}

		@Override
		public int hashCode()
		{	return mInstance.hashCode()^mZone.hashCode()^mIdentifier.hashCode();
		}
		
	}
}
